<?php
namespace Service;

use Doctrine\DBAL\Query\QueryBuilder;
use Doctrine\ORM\Query\ParameterTypeInferer;
use Doctrine\ORM\EntityManager;
use Util\MongoDataBaseAlterations;
use Util\Utilities;
use Zeedhi\Framework\DataSource\DataSet;
use Zeedhi\Framework\DataSource\FilterCriteria;
use \Service\DataSource;
use \Service\ImportSchedule;
use \Service\Manager\Mongo\Manager;
use Zeedhi\Framework\DependencyInjection\InstanceManager;
use Zeedhi\Zmart\DataSource\ZmartFilterCriteria;


class Indicator {

	protected $dataSourceName = 'indicator';
	/** @var \Service\Manager\Mongo\Manager $organizationManager */
	protected $organizationManagers;
	protected $importScheduleService;
	protected $persistedDataSources = array();

	protected $PARAMETERS_TO_CLEAN_BEFORE_IMPORT = array(
		'_id',
		'organizationName',
		'connection',
		'connectionName',
		'menu',
		'menuName'
	);

	/**
	 * This is the DataSourceService that does not use the mongo Manager
	 * @var DataSourceService
	 */
	protected $unmanagedDataSourceService;
	/** @var array */
	protected $parameterMapping = array(
		'period_start' => 'START_DATE',
		'period_end' => 'END_DATE',
		'filter' => 'STRUCTURE'
	);
	public static $INDICATOR_KEYS_TO_CLEAR = array("_id", "menu", "menuName");
	public static $DATASOURCE_KEYS_TO_CLEAR = array("_id", "relatedStructuresType");

	public function __construct(Manager $organizationManager, DataSource $dataSourceService, array $defaultOrgManagers, DataSourceService $unmanagedDataSourceService, ImportSchedule $importScheduleService) {
		$this->organizationManager = $organizationManager;
		$this->dataSourceService = $dataSourceService;
		$this->defaultOrgManagers = $defaultOrgManagers;
		$this->unmanagedDataSourceService = $unmanagedDataSourceService;
		$this->importScheduleService = $importScheduleService;
	}

	public function findByCriteria($columnName, $value, $operator = '=') {
		$criteria = array( array(
			"columnName" => $columnName,
			"value" => $value,
			"operator" => $operator
		));

		$filterCriteria = new FilterCriteria($this->dataSourceName, $criteria);
		return $this->organizationManager->findBy($filterCriteria)->getRows();
	}

	public function findById($indicator) {
		$indicator = $this->findByCriteria('_id', $indicator);
		return count($indicator) > 0 ? $indicator[0] : null;
	}

    public function findDataSourceByOldId($id) {
        $criteria = array( array(
            "columnName" => 'oldId',
            "value" => $id,
            "operator" => '='
        ));

        $filterCriteria = new FilterCriteria('dataSource', $criteria);
        $dataSource=  $this->organizationManager->findBy($filterCriteria)->getRows();
        return count($dataSource) > 0 ? $dataSource[0] : null;
    }

	public function setLog($dataBefore, $id) {
        $dataAfter = $this->findById($id);
        Utilities::setLog(array($dataBefore), array($dataAfter), $this->dataSourceName, $id);
	}

	public function setLogImportation($importedItems) {

        if(!empty($importedItems['indicator'])){
            foreach ($importedItems['indicator'] as $indicator) {
                $this->setLog(array(),$indicator['_id']);
            }
        }

        if(!empty($importedItems['widget'])){
            /** @var $widgetService \Service\Widget */
            $widgetService = InstanceManager::getInstance()->getService('\Service\Widget');
            foreach ($importedItems['widget'] as $widget) {
                $widgetService->setLog(array(),$widget['_id']);
            }
        }

        if(!empty($importedItems['datasource'])){
            $dataSourceService = InstanceManager::getInstance()->getService('\Service\DataSource');
            foreach ($importedItems['datasource'] as $datasource) {
                if(!empty($datasource['_id'])){
                    $dataSourceService->setLog(null,$datasource['_id']);
                }
            }
        }
    }

	public function findByOldId($indicator) {
		$indicator = $this->findByCriteria('oldId', $indicator);
		return count($indicator) > 0 ? $indicator[0] : null;
	}

	public function findByDataSourceId($dataSourceId) {
		return $this->findByCriteria('dataSource', $dataSourceId);
	}

	public function findByMenu($menuId) {
		return $this->findByCriteria('menu', $menuId, FilterCriteria::EQ);
	}

	public function findByIndicators($indicators) {
		return $this->findByCriteria('indicator', $indicators, FilterCriteria::EQ);
	}

	public function alreadyExistsOnDb($indicator, $returnRealId = false) {
		$filterCriteria = new FilterCriteria($this->dataSourceName, array(array("columnName"=>"oldId", "value" => $indicator, "operator" => "=")));
		$indicator = $this->organizationManager->findBy($filterCriteria)->getRows();
		if($returnRealId){
			return isset($indicator[0]) ? $indicator[0]["_id"] : null;
		}else{
			return count($indicator) > 0 ? true : false;
		}
	}

	public function checkViewsWithIndicator($indicators){
        $viewService = InstanceManager::getInstance()->getService('\Service\View');
        foreach($indicators as $indicator){
            $viewWithIndicator = $viewService->findByIndicator($indicator["_id"]);
            if(!empty($viewWithIndicator)){
                throw new \Exception(Exception::buildErrorJSON(
                    "CANNOT_DELETE_INDICATOR",
                    array(
                        "view" => implode(", ",array_column($viewWithIndicator,"name")),
                        "indicator" => $indicator["name"]
                    )
                ), 1);
            }
        }
    }

	public function getRelatedStructuresType($indicator) {
		$structuresType = array();
		$dataSource = $this->dataSourceService->findByIndicator($indicator);
		if ($dataSource && isset($dataSource['relatedStructuresType'])) {
			$structuresType = $dataSource['relatedStructuresType'];
		}
		return $structuresType;
	}
	public function getRelatedStructuresTypeByDataSource($dataSourceId) {
		$structuresType = array();
		$dataSource = $this->dataSourceService->findById($dataSourceId);
		if (!empty($dataSource[0]['relatedStructuresType'])) {
			$structuresType = $dataSource[0]['relatedStructuresType'];
		}
		return $structuresType;
	}

	/**
	 * If $withRelationsOnly is true, only the relatedStructureTypes that have at least one relation will be returned
	 * @param $indicator
	 * @param bool $withRelationsOnly
	 * @return array
	 */
	public function getRelatedStructuresTypeIds($indicator, $withRelationsOnly = false) {
		$structuresTypeIds = array();
		$structuresType = $this->getRelatedStructuresType($indicator);

        return $this->validRelatedStruturesType($withRelationsOnly, $structuresType, $structuresTypeIds);
    }

    public function getRelatedStructuresTypeIdsByDataSource($dataSourceId, $withRelationsOnly = false) {
        $structuresTypeIds = array();
        $structuresType = $this->getRelatedStructuresTypeByDataSource($dataSourceId);
        return $this->validRelatedStruturesType($withRelationsOnly, $structuresType, $structuresTypeIds);
    }


	protected function filterStructureTypesWithRelation($structureTypes) {
		$filtered = array();

		foreach ($structureTypes as $structureType) {
			if (!empty($structureType['relations'])) {
				$filtered[] = $structureType;
			}
		}

		return $filtered;
	}

	public function getDataFromMultipleOrgs(FilterCriteria $filterCriteria){
		$data = array();
		foreach ($this->defaultOrgManagers as $dbLabel=>$defaultOrgManager) {
			$rows = $defaultOrgManager->findBy($filterCriteria)->getRows();
			foreach ($rows as &$row) {
				$row["organizationName"] = $dbLabel;
				array_push($data, $row);
			}
		}
		return $data;
	}

	/*Import routine*/
	protected function buildFilterCondition($collectionName, $field, $operator, $value){
		$filterCriteria = new ZmartFilterCriteria($collectionName);
		$filterCriteria->addCondition($field, $operator, $value);
		return $filterCriteria;
	}

	public function getDataByFilterCriteria($collectionName, $filterCriteria){
		$filterCriteria->setDataSourceName($collectionName);
		$data = $this->getDataFromMultipleOrgs($filterCriteria);
		return $data;
	}

	public function checkDrillDownIndicators($indicatorsIds){
		$result = $this->recursiveDrillDownIndicators($indicatorsIds, $indicatorsIds);
		return array_values(array_diff($result,$indicatorsIds));
	}

	public function recursiveDrillDownIndicators($indicatorsIds, $drillDownIndicatorsIds){
		foreach($indicatorsIds as $indicatorId){
			$viewCriteria = $this->buildFilterCondition("widget", 'indicator', ZmartFilterCriteria::EQ, $indicatorId);
			$views = $this->getDataByFilterCriteria("widget", $viewCriteria);
			if(!empty($views)){
				foreach($views as $view){
					if(!empty($view["drillDownView"])){
						$drillDownViewCriteria = $this->buildFilterCondition("widget", '_id', ZmartFilterCriteria::EQ, $view["drillDownView"]);
						$drillDownView = $this->getDataByFilterCriteria("widget", $drillDownViewCriteria);
						if(!empty($drillDownView[0]) && !in_array($drillDownView[0]["indicator"],$drillDownIndicatorsIds)){
							array_push($drillDownIndicatorsIds,$drillDownView[0]["indicator"]);
							$drillDownIndicatorsIds = $this->recursiveDrillDownIndicators([$drillDownView[0]["indicator"]], $drillDownIndicatorsIds);
						}
					}
				}
			}
		}
		return $drillDownIndicatorsIds;
	}

	public function updateDataSourceColumns($indicatorId,$dataSourceColumns){
		$widgetService = InstanceManager::getInstance()->getService('\Service\Widget');
		$widgets = $widgetService->findWidgetsByIndicator($indicatorId);
		foreach($widgets as $widget){
			$widget["columns"] = $dataSourceColumns;
			$this->organizationManager->persist(new DataSet('widget', array($widget)));
		}
	}

	public function getIndicatorsViewsAndDatasources($indicatorsId, $fromOrganization = false) {
		$data = array();
		$indicatorData = array();
		$condition = $this->buildFilterCondition($this->dataSourceName, '_id', ZmartFilterCriteria::IN, $indicatorsId);
		$indicatorsToImport = $fromOrganization? $this->organizationManager->findBy($condition)->getRows() : $this->getDataByFilterCriteria($this->dataSourceName, $condition);
		foreach ($indicatorsToImport as $indicatorToImport) {
			$indicatorData['indicator'] = $indicatorToImport;

			if(isset($indicatorToImport['dataSource'])){
				$condition = $this->buildFilterCondition('dataSource', Manager::MONGO_KEY_COLUMN, ZmartFilterCriteria::EQ,$indicatorToImport['dataSource']);
				$indicatorDatasources = $fromOrganization? $this->organizationManager->findBy($condition)->getRows() : $this->getDataByFilterCriteria('dataSource',$condition);
				$indicatorData['dataSources'] = $indicatorDatasources;
			}else{
				$indicatorData['dataSources'] = [];
			}
			$condition = $this->buildFilterCondition("widget", "indicator", ZmartFilterCriteria::IN, $indicatorsId);
			$views = $fromOrganization ? $this->organizationManager->findBy($condition)->getRows() : $this->getDataByFilterCriteria('widget', $condition);
			$indicatorData['views'] =  $views;
			array_push($data, $indicatorData);
		}

		return $data;
	}

	public function validLogNewDataSources($originalDataSources){
        $dataSourceLogs = array();
        foreach ($originalDataSources as $originalDataSource) {
            if(!empty($originalDataSource['_id'])){
                $id = $originalDataSource['_id'];
                $dataSource = $this->findDataSourceByOldId($id);
                if(empty($dataSource) && !in_array($id,$dataSourceLogs)){
                    $dataSourceLogs[] = $id;
                }
            }
        }
        return $dataSourceLogs;
	}

	public function saveDefaultOrgIndicatorsAndDataSources($indicatorsId, $returnOldViewIds = false,$menuOldIdList = null, $importOptions = null, $structureTypesOldIdList = []){
		$indicatorsToImport = array();
		$viewsToImport = array();
		$persistedDataSourceIds = array();
        $dataSourceLogs = array();
		$indicatorsItems = $this->getIndicatorsViewsAndDatasources($indicatorsId);
        /** @var Widget $widgetService */
		$widgetService = InstanceManager::getInstance()->getService('\Service\Widget');
		$isImportDefaultOrg = Utilities::getOrganizationIdForMongo() == "5";
		foreach ($indicatorsItems as $i => $indicatorItems) {
			$indicator = $indicatorItems['indicator'];
			$dataSources = $indicatorItems['dataSources'];
			$views = $indicatorItems['views'];


			foreach($views as &$view){
				if($isImportDefaultOrg){
					$view["mockPeriod"] = empty($view["period"]) ? null : $widgetService->setPeriodFormat($view["period"]);
				}
				$this->importPeriod($view);
			}


			$indicator['oldId'] = $indicator['_id'];

			$removeMenu = true;
			if( !empty($menuOldIdList) && !empty($indicator["menu"])
				&& !empty($menuOldIdList[$indicator["menu"]])
				&& $importOptions != "dashboardOnly"){
				$indicator["menu"] = $menuOldIdList[$indicator["menu"]];
				$removeMenu = false;
			}

			$this->cleanBeforeImport($indicator,$removeMenu);

            $dataSourceLogsValid = $this->validLogNewDataSources($dataSources);
			$persistedDataSource = $this->importIndicatorsDatasources($dataSources,$isImportDefaultOrg,$structureTypesOldIdList);
			$indicator['dataSource'] = $persistedDataSource[0]["_id"];

            if(!empty($dataSourceLogsValid)){
                $dataSourceLogs[] = $persistedDataSource[0];
            }

			array_push($persistedDataSourceIds, $persistedDataSource);
			array_push($viewsToImport, $views);
			array_push($indicatorsToImport, $indicator);
		}

		$persistedIndicatorsIds = $this->organizationManager->persist(new DataSet('indicator', $indicatorsToImport));

		$persistedItems = array();
		if(sizeof($persistedIndicatorsIds)>0){
			$persistedItems['indicator'] = $persistedIndicatorsIds;
			$viewsIds = $this->importIndicatorsViews($persistedIndicatorsIds, $viewsToImport, true);
			$persistedItems['widget'] = $returnOldViewIds ? $viewsIds : $viewsIds[0];
			$persistedItems['datasource'] = $persistedDataSourceIds;
			$persistedItems['newDatasource'] = $dataSourceLogs;
		}
		return $persistedItems;
	}

	public function saveDefaultOrgPeriod($period){
		unset($period["_id"]);
		$savedPeriod = $this->organizationManager->persist(new DataSet('periods', [$period]));
		if(!empty($savedPeriod) && !empty($savedPeriod[0]) && !empty($savedPeriod[0]["_id"])){
			return $savedPeriod[0]["_id"];
		}
		return null;
	}

	protected function getPeriodFilterCriteriaByPeriodCode($periodCode){
		return new FilterCriteria("periods",
			array(array("columnName"=>"periodsCode", "value" => $periodCode, "operator" => "="))
		);
	}

	protected function importSavePeriod(&$row){
		if(!empty($row) && !is_array($row)){
			try{
				$condition = $this->buildFilterCondition("periods", '_id', ZmartFilterCriteria::EQ, $row);
				$period = $this->getDataByFilterCriteria('periods',$condition);
			}catch (\Exception $e){

			}
			if(!empty($period) && is_array($period) && !empty($period[0])){
				$period = $period[0];
			}else{
				$period = $this->getDataFromMultipleOrgs($this->getPeriodFilterCriteriaByPeriodCode($row))[0];
			}
			if(!empty($period["periodsCode"])){
				$periodOnDb = $this->organizationManager->findBy($this->getPeriodFilterCriteriaByPeriodCode($period["periodsCode"]))->getRows();
				if(empty($periodOnDb) && empty($periodOnDb[0])){
					$this->saveDefaultOrgPeriod($period);
					$row = $period["periodsCode"];
				}else{
					$row = $periodOnDb[0]["periodsCode"] ?: null;
				}
			}else if(!empty($period["START_DATE_EXPRESSION"]) || !empty($period["END_DATE_EXPRESSION"])){
				$condition = new FilterCriteria("periods",
					array(array("columnName"=>"START_DATE_EXPRESSION", "value" => $period["START_DATE_EXPRESSION"] ?: "", "operator" => "="))
				);
				$condition->addCondition("END_DATE_EXPRESSION","=",$period["END_DATE_EXPRESSION"] ?: "");
				$periodOnDb = $this->organizationManager->findBy($condition)->getRows();
				if(empty($periodOnDb) && empty($periodOnDb[0])){
					$row = $this->saveDefaultOrgPeriod($period);
				}else{
					$row = $periodOnDb[0]["_id"];
				}
			}else{
				$row = null;
			}
		}
	}

	public function importPeriod(&$row){
		try{
			if(is_array($row["period"])){
				foreach($row["period"] as &$period){
					try{
						$this->importSavePeriod($period);
					}catch (\Exception $e){

					}
				}
			}else{
				$this->importSavePeriod($row["period"]);
			}
		}catch (\Exception $e){

		}
	}

	public function importIndicatorsDatasources($originalDataSources, $isImportDefaultOrg = false,$structureTypesOldIdList=[]){
		foreach ($originalDataSources as $originalDataSource) {
			$originalDataSourceId = $originalDataSource['_id'];

			$newDataSource = $originalDataSource;
			$newDataSource['oldId'] = $originalDataSourceId;
			$dataSource = $this->removeObjectsAlreadyImported(array($newDataSource),'dataSource');

			if(isset($dataSource[0])){
				$newDataSource = $dataSource[0];
			}else{
				return $this->removeObjectsAlreadyImported(array($newDataSource),'dataSource',true);
			}

			$newDataSource["isScheduleActive"] = !$isImportDefaultOrg;
			if(!$isImportDefaultOrg){
				$newDataSource["modified"] = false;
			}

			$this->cleanBeforeImport($newDataSource);

			list($newDataSource["connection"], $newDataSource["connectionName"]) = $this->getMainConnection();

			$relatedStructuresType = $this->unmanagedDataSourceService->mergeDefaultRelations($newDataSource,$structureTypesOldIdList);
			$newDataSource['relatedStructuresType'] = $relatedStructuresType;

			$this->importPeriod($newDataSource);

			$persistedDataSourceId = $this->organizationManager->persist(new DataSet('dataSource', array($newDataSource)));
			$persistedDataSources[$originalDataSourceId] = $persistedDataSourceId;

			if($isImportDefaultOrg){
				$this->organizationManager->cloneCollectionToImportDatabase($originalDataSourceId, $persistedDataSourceId[0]["_id"]);
			}

			if(!empty($originalDataSource["dateFilterColumn"])){
				$this->dataSourceService->saveIndexes($persistedDataSourceId[0]["_id"], $originalDataSource["dateFilterColumn"], [$originalDataSource["dateFilterColumn"]], [1]);
				$this->dataSourceService->saveIndexes($persistedDataSourceId[0]["_id"], $originalDataSource["dateFilterColumn"]."_DESC", [$originalDataSource["dateFilterColumn"]], [-1]);
			}

			if(!empty($newDataSource["scheduleExpression"]) && !empty($newDataSource["isScheduleActive"])){
				$this->importScheduleService->saveScheduleExpression($newDataSource, $persistedDataSourceId);
			}

			return $persistedDataSourceId;
		}
	}

	public function getMainConnection(){
		$connectionsAlreadyOnDb = $this->organizationManager->findBy(new FilterCriteria("connection",
			array(array("columnName"=>"main", "value" => true, "operator" => "="))
		))->getRows();
		if(!empty($connectionsAlreadyOnDb) && !empty($connectionsAlreadyOnDb[0])){
			$connection = $connectionsAlreadyOnDb[0]["_id"];
			$connectionName = $connectionsAlreadyOnDb[0]["name"];
		} else {
			$connectionsAlreadyOnDb = $this->organizationManager->findBy(new FilterCriteria("connection",array()))->getRows();
			if(!empty($connectionsAlreadyOnDb) && !empty($connectionsAlreadyOnDb[0])){
				$connection = $connectionsAlreadyOnDb[0]["_id"];
				$connectionName = $connectionsAlreadyOnDb[0]["name"];
			}
		}
		return array($connection, $connectionName);
	}

	protected function removeObjectsAlreadyImported($toInsert,$dataSourceName,$returnIdsOnEmptyReturn = false){
		$backup = array();
		$alreadyOnDb = $this->organizationManager->findBy(new FilterCriteria($dataSourceName,array()))->getRows();
		foreach($toInsert as $key=>$item){
			foreach($alreadyOnDb as $dbObject){
				if((isset($dbObject['oldId']) && isset($item['_id']) && $item['_id'] === $dbObject['oldId']) ||
					(isset($dbObject['oldId']) && $item['oldId'] === $dbObject['oldId'])){
					array_push($backup, $dbObject);
					unset($toInsert[$key]);
				}
			}
		}

		return $returnIdsOnEmptyReturn && !isset($toInsert[0]) ? $backup : $toInsert;
	}
    public static function getDrillDownViewImport(array $array,$value,&$view){
        foreach ($array as $key => $row){
            if(!empty($row["oldId"]) && $value == (string)$row['oldId']){
                return (string)$row['_id'];
            }
		}
		$view["hasDrillDown"] = false;
		$view["drillDownViewField"] = "";
        return null;
    }

	public function setAllDrillDowns($importedItems){
		$indicatorsIds=[];
		foreach($importedItems["widget"] as $indicator){
			if(!empty($indicator["_id"]) && !in_array($indicator["_id"],$indicatorsIds)){
				array_push($indicatorsIds,$indicator["_id"]);
			}
		}
		$condition = $this->buildFilterCondition("widget", '_id', ZmartFilterCriteria::IN, $indicatorsIds);
		$allViews = $this->organizationManager->findBy($condition)->getRows();
		foreach($allViews as &$view){
			if(!empty($view['drillDownView'])){
				$view['drillDownView'] = self::getDrillDownViewImport($allViews,$view['drillDownView'],$view);
			}
		}
		$this->organizationManager->persist(new DataSet('widget', $allViews));
	}

	public function importIndicatorsViews($persistedIndicatorsIds, $indicatorsViews, $returnOldId = false){
		$newViews = array();
		$oldIds = [];
		foreach ($indicatorsViews as $index=>&$views) {
			foreach ($views as &$view) {
				$view['indicator'] = $persistedIndicatorsIds[$index]['_id'];
				$view['indicators']['indicator'] = $persistedIndicatorsIds[$index]['_id'];
				$view['oldId'] = $view['_id'];
				array_push($oldIds, $view['oldId']);
				$this->cleanBeforeImport($view);
				array_push($newViews, $view);
			}
		}
		$persistedViewsIds = $this->organizationManager->persist(new DataSet('widget', $newViews));
        $return = $returnOldId ? array($persistedViewsIds, $oldIds) : $persistedViewsIds;
		return $return;
	}

	public function saveDefaultOrgViews($widgets){
		return $this->organizationManager->persist(new DataSet('widget', $widgets));
	}

	protected function cleanBeforeImport(&$object, $removeMenu = true){
		foreach($this->PARAMETERS_TO_CLEAN_BEFORE_IMPORT as $parameter){
			if(isset($object[$parameter]) && !(!$removeMenu && in_array($parameter, ['menu', 'menuName']))){
				unset($object[$parameter]);
			}
		}
	}

    /**
     * @param $withRelationsOnly
     * @param array $structuresType
     * @param array $structuresTypeIds
     * @return array
     */
    public function validRelatedStruturesType($withRelationsOnly, array $structuresType, array $structuresTypeIds): array
    {
        if ($withRelationsOnly) {
            $structuresType = $this->filterStructureTypesWithRelation($structuresType);
        }

        foreach ($structuresType as $structureType) {
            $structuresTypeIds[] = $structureType['structureType'];
        }
        return $structuresTypeIds;
    }

}
