<?php
namespace Zeedhi\Report;

/**
 * Description of Loader
 *
 * @author tuliopinto
 */
class Loader {


    private $datasetQueryLoader;

    /**
     * Constructor.
     *
     * @param \Zeedhi\Report\Loader\Dataset\DatasetLoadable $datasetQueryLoader
     */
    public function __construct( Loader\Dataset\DatasetLoadable $datasetQueryLoader) {
        $this->datasetQueryLoader = $datasetQueryLoader;
    }

    /**
     * Build report from report name.
     *
     * @param string $reportName
     * @param LogoConfig $logoConfig
     *
     * @return string Generated html file path.
     */
    public function buildReport($reportName, LogoConfig $logoConfig, $entityManager) {
        $ds = DIRECTORY_SEPARATOR;
        $templatePath = dirname(__FILE__) . $ds . "templates";
        $cachePath = $templatePath . $ds . "cache";

        $targetFilePath = $this->createTargetFilePath($reportName);
        $reportMetadata = $this->loadReportMetadata($reportName);
        $reportMetadataSource = $reportMetadata['dataSource'];
        $dataSourceName = $reportMetadataSource['name'];
        if($dataSourceName)
        {
            $reportData = $this->loadReportData($reportMetadata, $entityManager);
        }else{
            $reportData = $reportMetadataSource['data'];
        }


        $reportBuilder = new \Zeedhi\Report\ReportBuilder($reportMetadata, $reportData, $logoConfig, $templatePath, $cachePath, $targetFilePath);
        $reportBuilder->buildReport();
        return $targetFilePath;
    }


    public function loadReportMetadata($reportName) {
        $jsonFileName = strtolower($reportName).'.json';
        $ds = DIRECTORY_SEPARATOR;
        $jsonFilePath = dirname(__NAMESPACE__). $ds . ".." . $ds . "/../gen/reports/$jsonFileName";
        $reportMetadata  = json_decode(file_get_contents($jsonFilePath), true);
        return $reportMetadata;
    }

    /**
     * Create target file path from report name.
     *
     * @param string $reportName
     *
     * @return string
     */
    public function createTargetFilePath($reportName) {
        $ds = DIRECTORY_SEPARATOR;
        $htmlFileName = strtolower($reportName).'.html';
        $targetFilePath = dirname(__NAMESPACE__). $ds . ".." . $ds ."/../gen/reports/$htmlFileName";
        return $targetFilePath;
    }

    /**
     * Create target file url from report name.
     *
     * @param string $reportName
     *
     * @return string
     */
    public function createTargetFileUrl($reportName) {
        $htmlFileName = strtolower($reportName).'.html';
        return "../backend/gen/reports/$htmlFileName";
    }

    /**
     * Load report data from reportName.
     *
     * @param string $reportName
     *
     * @return array
     */
    public function loadReportData($reportName, $entityManager) {
        $datasetQueryLoader = $this->getDatasetQueryLoader();
        $sql = $datasetQueryLoader->loadReportSql($reportName);

        $queryBuilder = $entityManager->createQueryBuilder();
        $params = array();
        try{
            $queryBuilder->select('wpv.value')
                ->from('\Zeedhi\Model\Viewer\Entities\ZhWidgetPropertyValue', 'wpv')
                ->join('wpv.widget', 'w')
                ->join('wpv.templateProperty', 'tp')
                ->where('w.name = :widgetName')
                ->setParameter('widgetName', $reportName);

            $queryBuilderParams = clone $queryBuilder;
            $queryBuilderParams->andWhere('tp.name = :propertyName')
                ->setParameter('propertyName', 'sqlParams');
            $queryParams = $queryBuilderParams->getQuery();
            $resultParams = $queryParams->getResult();
            if (!empty($resultParams) && $resultParams[0]['value'] && $temp = json_decode($resultParams[0]['value'], true)) {
                $params = $temp;
            }
        }catch (\Exception $e){
            //$params = array(); @todo warning
        }
        $connection= $entityManager->getConnection();
        $data = $connection->fetchAll($sql, $params);
        return $data;
    }

    /**
     * Create a connection object from a connection entity;
     *
     * @param \Zeedhi\Model\DataBase\Entities\ZhConnection $connectionEntity
     *
     * @return \Doctrine\DBAL\Connection
     */
    private function createConnectionFromEntity(\Zeedhi\Model\DataBase\Entities\ZhConnection $connectionEntity) {
        $connectionParams = array(
            'host'     => $connectionEntity->getHost(),
            'port'     => $connectionEntity->getPort(),
            'dbname'   => $connectionEntity->getDbname(),
            'user'     => $connectionEntity->getUsername(),
            'password' => $connectionEntity->getPassword(),
            'driver'   => $connectionEntity->getDriver()
        );

        $eventManager = new \Doctrine\Common\EventManager();
        $eventManager->addEventSubscriber(new \Doctrine\DBAL\Event\Listeners\OracleSessionInit());

        $conn = \Doctrine\DBAL\DriverManager::getConnection($connectionParams, new \Doctrine\DBAL\Configuration(), $eventManager);
        return $conn;
    }

    /**
     *
     *
     * @param string $reportName
     *
     * @return object
     */
    public function createReportDataSourceObject($dataSourceName) {
        $namespace = '\\Generated\\Datasource\\';
        $className = \Doctrine\Common\Util\Inflector::classify(strtolower($dataSourceName));
        $fullClassName = $namespace . $className;
        return $fullClassName;
    }

    /**
     *
     * @return \Zeedhi\Framework\Application
     */
    public function getApplication() {
        return $this->application;
    }

    public function getDatasetQueryLoader() {
        return $this->datasetQueryLoader;
    }

}